import os

import math
import numpy as np

import pandas as pd

from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import scale
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA

import pickle
import re

from siVAE import util


def save_as_npy(dir_data, dir_label):

    dir_data_suffix, filetype = dir_data.rsplit(".",1)

    if filetype == 'csv':
        df_in = pd.read_csv(os.path.join(dir_data), index_col = 0)
        data = df_in.values.transpose().astype('float32')
        labels_gene = np.array(list(df_in.index))
        labels_type = np.array([id.split("_")[0] for id in list(df_in.columns.values)])
    else:
        raise Exception("File type must be csv, not {}".format(filetype))

    if dir_label is not None:
        labels_type = np.array(pd.read_csv(dir_label)).reshape(-1).tolist()

    np.savez(filename, matrix = data, column =  labels_type, index =labels_gene)
    # data = [gene,type]
    assert len(labels_type) == data.shape[0], 'Length of cell label ({}) does not match number of cells ({})'.format(len(labels_type), data.shape[0])


def scale_float64(data, *args, **kwargs):
    """
    Convert the input data to float64 to avoid errors
    """
    dtype = data.dtype
    data_scaled = scale(np.array(data).astype('float64'), *args, **kwargs)
    return data_scaled.astype(dtype)


def index2mask(index, length = None):
    if length is None:
        length = np.max(index) + 1
    return np.array([np.isin(np.arange(length), idx) for idx in index])


def extract_targets_from_labels(targets, labels):
    """
    targets = list of targets
    labels = numpy array of labels
    """
    if targets == ["all"]:
        indices = np.repeat(True, len(labels))
    else:
        indices = np.array([np.any([bool(re.search(target + "(_|\Z)", label)) for target in targets]) for label in labels])
    return indices


def import_labeled_df(dir_data, dir_label = None):
    """ Import data """

    datah = data_handler()
    datah.import_X(dir_data, dir_label)
    data = datah.data
    labels_gene = datah.labels_gene
    labels_cell = datah.labels_cell
    return (data, labels_gene, labels_cell)


def save_mask_gene(filename, matrix, column, index, format = "all"):
    """
    matrix = num_gene x num_pathway
    column: pathway
    index: gene
    """
    np.savez(filename, matrix = matrix, column = column, index = index)


def save_X(matrix, column, index, format = "all"):
    """
    matrix = num_samples x num_genes
    column: gene
    index: celltype
    """
    np.savez(filename, matrix = matrix, column = column, index = index)



def kfold_split(X,y, k_split, random_seed = 1, isPrecompMatrix = False):
    skf = StratifiedKFold(n_splits     = k_split,
                          random_state = random_seed,
                          shuffle      = True)
    if not isPrecompMatrix:
        datasets = [[X[idx_train],
                     X[idx_test],
                     y[idx_train],
                     y[idx_test]] for idx_train, idx_test in skf.split(X, y)]
    else:
        datasets = [[X[idx_train][:,idx_train],
                     X[idx_test][:,idx_train],
                     y[idx_train],
                     y[idx_test]] for idx_train, idx_test in skf.split(X, y)]
    return datasets


class data_handler(object):

    """
    Main Data (X, y, masks)
    Batch split
    Kfold split
    Dataset
    """

    def __init__(self, h_dims = None, X = None, y = None, batch_type = None,
                 masks = None, index_latent_embedding = 0, label = None,
                 idx_selected = np.array([], dtype = 'int'), use_3D = False,
                 masking = False, dir_y = None, scale_feature = True, data = None,
                 iterator = None):

        self.X = X
        self.y = y
        self.masks = masks
        self.label = label
        self.idx_selected = idx_selected
        self.batch_type = batch_type
        self.use_3D = use_3D
        self.masking = masking
        self.dir_y = dir_y
        self.data = None

        self.iterator = iterator

        self.index_latent_embedding = index_latent_embedding
        self.scale_feature = scale_feature

        self.set_h_dims(h_dims)


    def set_batch_type(self, batch_type):
        self.batch_type = batch_type


    def get_batch_mask(self):

        if self.batch_type == 'cell':
            self.batch_mask = self.mask_cell[self.index_cell_use]

        elif self.batch_type == 'gene':
            self.batch_mask = self.mask_gene[self.index_gene_use]

        else:
            raise Exception('Batch type must be either cell or gene')

        return self.batch_mask


    def set_batch_mask(self, batch_mask):
        self.batch_mask = batch_mask


    def get_kfold_index(self, kfold_idx):
        return self.kfold_index_list[kfold_idx]


    def set_use_3D(self, use_3D):
        self.use_3D = use_3D


    def set_use_masking(self, use_masking):
        self.use_masking = use_masking


    def create_dataset(self, kfold_idx = None):
        """
        split_idx: index of the training and test set in X and y data
        """

        if kfold_idx is None:
            split_idx = self.split_idx
        else:
            split_idx = self.get_kfold_index(kfold_idx)
            self.split_idx = split_idx

        data_list = [self.X, self.y]
        self.dataset = split_data(data_list, split_idx)


    def create_masks_batch(self, run_idx = None):
        masks_selected = self.masks[self.idx_selected]
        masks_combined = self.masks + np.any(masks_selected, axis = 0)
        self.masks_batch = masks_combined

        if run_idx is not None:
            self.masks_batch = masks_combined[self.idx_list[run_idx]]


    def create_X_batch(self, masking = False, scale_feature = False, use_3D = False):
        """ """
        if use_3D and not np.all(self.masks_batch.sum(1) == self.masks_batch.sum(1)[0]):
            raise Exception('For use_3D, all masks should contain same number of non-masked value')

        if masking:
            self.X_batch = np.array([self.X])
        else:
            self.X_batch = np.array([self.X[:,mask] for mask in self.masks_batch])

        if scale_feature:
            self.X_batch = [scale_float64(data_masked,0) for data_masked in self.X_batch]

        if not use_3D:
            assert len(data) == 1, "For use_3D is false, length of masks should be 1"
            self.X_batch = self.X_batch[0]


    def extract_batch(self, batch_num, batch_idx):
        idx_all_masks = np.arange(self.masks.shape[0])
        idx_unused = idx_all_masks[np.logical_not(np.isin(idx_all_masks, self.idx_selected))]
        idx_batch = np.array_split(idx_unused, batch_num)[batch_idx]

        self.idx_batch = idx_batch


    def check_rewrite(self, logdir = '', rewrite = False):
        """ """
        num_sel = len(self.idx_selected)

        if rewrite:
            idx_batch_incomplete = self.idx_batch

        else:
            idx_batch_incomplete = []
            for ii in self.idx_batch:
                filename = os.path.join(logdir, "sel-{}_gene-{}.npy".format(num_sel,ii))
                if not os.path.exists(filename):
                    idx_batch_incomplete.append(ii)

        self.idx_batch = np.array(idx_batch_incomplete)


    def prepare_batch(self, batch_num, batch_idx, n_nn, logdir, rewrite = False):
        """
        Prepare list of mask by extracting the (batch_idx) th batch out of
        (batch_num batches) then dividing that so the batches have n_nn masks
        """
        # Prepare idx lists
        self.extract_batch(batch_num, batch_idx)
        self.check_rewrite(logdir = logdir, rewrite = rewrite)

        if len(self.idx_batch) == 0:
            raise Exception('All masks in the batch have been completed.')

        idx_list = np.array_split(self.idx_batch, self.num_loops)

        self.num_loops = int(math.ceil(len(self.idx_batch)*1.0/n_nn))
        self.idx_list = idx_list


    def get_num_batch(self):
        return self.num_loops


    def import_X(self, dir_data, dir_label = None, dir_label_gene = None):
        """ Import data """
        filetype = dir_data.split('.')[-1]

        if filetype == 'npz':

            npz_in = np.load(dir_data, allow_pickle = True)
            data        = npz_in['matrix']
            labels_type = npz_in['index']
            labels_gene = npz_in['column']

        else:
            if filetype == 'csv':
                df_in = pd.read_csv(os.path.join(dir_data), index_col = 0)
                data = df_in.values.transpose().astype('float32')
                labels_gene = np.array(list(df_in.index))
                labels_type = [id.split("_")[0] for id in list(df_in.columns.values)]

            elif filetype == 'npy':
                data = np.load(dir_data, allow_pickle = True).astype('float32')

                if dir_label is None or dir_label_gene is None:
                    raise Exception('If input dir_data is npy format, dir_label must be specified')
            else:
                raise Exception('dir_data ({}) does not have correct format (csv,npy)'.format(dir_data))

            if dir_label is not None:
                labels_type = np.array(pd.read_csv(dir_label)).reshape(-1).tolist()

        assert len(labels_type) == data.shape[0], 'Length of cell label ({}) does not match number of cells ({})'.format(len(labels_type), data.shape[0])

        self.data = data # num_sample x num_gene
        self.labels_gene = labels_gene
        self.labels_cell = labels_type

        if self.output_dim == 0:
            self.set_h_dims(self.h_dims)

        # return (data, labels_gene, labels_type)


    def set_index_cell(self, dir_sample = "None", index_cell = None, targets = None):

        if dir_sample == "None":

            if index_cell is not None:
                self.index_cell = index_cell

            elif targets is not None:
                self.index_cell = []
                self.targets = targets

                # assert  np.all(np.all(np.isin(targets, list(set(self.labels_cell))))) != 0, 'Target ({}) not available in dataset'.format(targets)

                index_cell = extract_targets_from_labels(targets, self.labels_cell)

                if (targets != ['all']):
                    self.index_cell.append(index_cell) # Find indices for targets
                else:
                    self.index_cell.append(index_cell)

            else:
                raise Exception('Either index cell or targets needs to be specified')

        else:
            self.index_cell = np.load(dir_sample, allow_pickle = True)

        self.index_cell = np.array(self.index_cell)
        self.mask_cell = index2mask(self.index_cell, len(self.labels_cell))
        self.num_cell = len(self.index_cell[0])
        self.labels_target = np.zeros([1,self.num_cell])

        self.use_index_cell(index = "all")


    def set_split_idx(self, kk):
        self.split_idx = self.kfold_index_list[kk]


    def create_kfold_index_list(self, k_split, random_seed = 0):

        self.kfold_index_list = split_index(length = self.X.shape[-2],
                                            k_split = k_split,
                                            label =self.labels_target[0],
                                            random_seed = random_seed)


    def set_index_gene(self, index_gene = None, mask_gene = None):
        if index_cell is not None:
            self.index_gene = index_gene
            self.mask_gene = index2mask(self.index_gene)

        elif mask_cell is not None:
            self.index_gene = [np.where(mask) for mask in mask_gene]
            self.mask_gene = mask_gene

        self.use_index_gene('all')


    def use_index_cell(self, index = 'all'):
        if index == 'all':
            self.index_cell_use = np.arange(len(self.index_cell))
        else:
            self.index_cell_use = index
            self.index_cell = self.index_cell[index]


    def use_index_gene(self, index = 'all'):
        if index == 'all':
            self.index_gene_use = np.arange(len(self.index_gene))
        else:
            self.index_gene_use = index


    def filter_cell(self, index_cell = None, scale_feature = None, scale_axis = []):
        # Extract Targets
        self.data_target = []
        self.labels_target = []

        if scale_feature is None:
            scale_feature = self.scale_feature

        if index_cell is None:
            index_cell = self.index_cell

        print("index_cell: {}".format(index_cell.shape))

        # index_cell = index_cell[self.index_cell_use]

        for index in index_cell:
            data_target = self.data[index]
            labels_target = np.array(self.labels_cell)[index]

            # Normalize Target Data
            for ii in scale_axis:
                data_target = scale_float64(data_target, ii)

            if scale_feature:
                data_target = scale_float64(data_target, 0)

            self.data_target.append(data_target)
            self.labels_target.append(labels_target)

        self.data_target = np.array(self.data_target)
        self.label_target = np.array(labels_target)


    def filter_gene(self, mask_gene = None):
        if mask_gene is None:
            print("index_gene: {}".format(self.index_gene.shape))
            print("index_gene_use: {}".format(self.index_gene_use))
            index_gene = self.mask_gene[self.index_gene_use]
            mask_gene = index2mask(index_gene, len(self.labels_gene))
            print("mask_gene: {}".format(self.mask_gene.shape))
            # mask_gene = self.mask_gene[self.index_gene_use]
        print("mask_gene: {}".format(mask_gene.shape))
        print("data_target: {}".format(self.data_target.shape))
        # data_target = np.array([[data_temp.take(index, axis = -1) for index in index_gene] for data_temp in self.data_target])
        data_target = np.array([filter_gene(data_temp, masks = mask_gene, use_3D = self.use_3D, masking = self.masking) for data_temp in self.data_target])

        self.data_target = data_target


    def import_mask_gene(self, dir_pathway, labels_gene = None, match_gene = True, min_size = 10, max_size = 500):
        """ Import and process pathway in matrix form """

        if labels_gene is None:
            labels_gene = self.labels_gene

        if dir_pathway == 'all_genes':
            mat_pathway = np.repeat(True, len(labels_gene)).reshape(1,-1)
            label_pathway = np.array(['all_genes'])

        elif dir_pathway == "single_genes":
            mat_pathway = np.identity(len(labels_gene), dtype = 'bool')
            label_pathway = labels_gene

        else:
            filetype = dir_pathway.split('.')[-1]

            if 'single' in dir_pathway and min_size != 1:
                min_size = 1
                print('For single genes, min_size has been set to 1.')

            if filetype == 'csv':
                df_pathway = pd.read_csv(dir_pathway, index_col = 0)
                pathway_gene = np.array(df_pathway.index)
                label_pathway = df_pathway.columns.values

            elif filetype == 'npz':
                npz_file = np.load(dir_pathway, allow_pickle = True)

                mat_pathway   = npz_file['matrix'].astype('bool') # load n_path x n_index
                label_pathway = npz_file['column']
                pathway_gene  = npz_file['index']

                df_pathway = pd.DataFrame(mat_pathway)
                df_pathway.columns = label_pathway
                df_pathway.index = pathway_gene

            else:
                raise Exception('File type must be either csv, npy, npz or specified as all_genes/single_genes')

            # Filter and reindex genes in the imported mat_pathway to match the expression matrix
            if labels_gene is not None:
                if match_gene and not np.all(np.isin(labels_gene, pathway_gene)):
                    raise Exception('Gene in Expression Matrix and Pathways do not match')
                else:
                    df_pathway = df_pathway.reindex(labels_gene).fillna(0)

            mat_pathway = df_pathway.values.transpose().astype('bool')

            # Filter out pathways that does not meet the min_size
            idx_pathway = np.all([mat_pathway.sum(1) >= min_size, mat_pathway.sum(1) <= max_size], axis = 0)

            num_sizeerror = sum(~idx_pathway)

            if num_sizeerror > 0:
                print('{} / {} did not mean size requirement'.format(num_sizeerror,
                                                                     len(label_pathway)))
                mat_pathway = mat_pathway[idx_pathway]
                label_pathway = label_pathway[idx_pathway]

        # Set outputs
        self.mask_gene = mat_pathway
        self.index_gene = np.array([np.where(mask)[0] for mask in self.mask_gene])
        self.label_geneset = label_pathway
        self.use_index_gene("all")


    def set_batch_index(self, batch_index):
        self.batch_index = batch_index


    def set_h_dims(self, h_dims):
        """ Set h_dims of data_handler object """
        self.h_dims = h_dims

        if self.h_dims[-1] == 0:
            if self.data is not None:
                self.h_dims[-1] = self.data.shape[-1] # set h_dims to be number of genes

        self.output_dim = h_dims[-1]

    def import_y(self, dir_y, index_cell = None, with_std = False, target_specific_PCA = True):
        print("importing from dir_y = {}".format(dir_y))
        if dir_y == 'None':

            kwargs = {"dir_y": dir_y,
                      "y_dim": self.h_dims[-1],
                      "data": self.data,
                      "with_std": with_std,
                      "target_specific_PCA": target_specific_PCA
                      }

            if index_cell is None:
                index_cell = self.index_cell

            self.y = np.array([import_y(index_targets = index, **kwargs) for index in index_cell])

        elif dir_y == 'X':
            self.y = self.X

        else:
            self.y = import_y(dir_y)
            if len(self.y.shape) == 3:
                if self.batch_type == 'cell':
                    self.y = self.y.take(self.batch_index, axis = 0)
                    assert len(self.y) == len(self.index_cell)
            else:
                self.y = np.array([self.y])



    def prepare_for_batch(self, dir_y = None):

        if self.batch_type == 'gene':
            self.filter_cell()
            if dir_y is None:
                dir_y = self.dir_y
            self.import_y(dir_y)


    def prepare_for_CV(self, dir_y):
        if self.batch_type == 'cell':
            self.filter_cell()
            self.import_y(dir_y)


    def prepare_X(self):

        if self.batch_type == 'cell':
            self.filter_cell(self.batch_mask)
            print("Data_target after filter cell - {}".format(self.data_target.shape))
            self.filter_gene()
            print("Data_target after filter gene - {}".format(self.data_target.shape))

        elif self.batch_type == 'gene':
            self.filter_cell()
            print("Data_target after filter cell - {}".format(self.data_target.shape))
            self.filter_gene(self.batch_mask)
            print("Data_target after filter gene - {}".format(self.data_target.shape))

        self.X = self.data_target.reshape(-1, *self.data_target.shape[2:])


    def prepare_y(self, dir_y = None):

        if dir_y is None:
            dir_y = self.dir_y

        print("dir_y = {}".format(dir_y))

        if dir_y == 'X':
            self.y = self.X
        else:
            if self.batch_type == 'cell':
                self.import_y(dir_y, index_cell = self.batch_mask)


    def prepare_CV(idx, masks_in, n_nn):
        datah.set_batch_index(idx)
        datah.set_batch_mask(masks_in)
        datah.n_nn = n_nn
        datah.prepare_X()
        datah.prepare_y()


def preprocess_pca(data, n_pca, whiten = False, verbose = False, random_state = 0, center = True, with_std = False):
    """
    data = [sample,feature]
    """

    # Convert to float64 to avoid numerical error warning
    dtype = data.dtype
    # data_scaled = scale_float64(data,0)
    data_scaled = data

    if data.shape[-1] == n_pca:
        data_pca = data_scaled

    else:
        pca = PCA(n_components=n_pca, whiten = False, random_state = random_state)
        data_pca = pca.fit_transform(data_scaled)

        if center:
            data_pca = scale_float64(data_pca, 0, with_std = with_std)

    if verbose:
        print('Explained Variance: {}'.format(sum(pca.explained_variance_ratio_)))

    return data_pca


def import_y(dir_y, y_dim = None, data = None, index_targets = None, with_std = False, target_specific_PCA = True):
    """
    """
    if dir_y == 'None':

        if data is None or index_targets is None or y_dim is None:
            raise Exception('Data, y_dim, and index_targets required')

        else:
            n_pca = y_dim

            if n_pca == data.shape[-1]:
                data_pca = data[index_targets]

            else:

                if target_specific_PCA:
                    data_pca = preprocess_pca(data[index_targets], n_pca, whiten = False, verbose = True)

                else:
                    data_pca_full = preprocess_pca(data, n_pca, whiten = False, verbose = True)
                    data_pca = data_pca_full[index_targets]

        data_pca = scale_float64(data_pca, 0, with_std = with_std)

    else:
        filetype = dir_y.split('.')[-1]
        if filetype == 'csv':
            data_pca = pd.read_csv(dir_y, header = None).values.astype('float32')
        elif filetype == 'npy':
            data_pca = np.load(dir_y, allow_pickle = True).astype('float32')

        if len(data_pca.shape) == 2:
            data_pca = data_pca.reshape(-1, *data_pca.shape)

        data_pca = np.array([scale_float64(data_loop, 0, with_std = with_std) for data_loop in data_pca])

    return data_pca


def split_index(length, k_split, label = None, random_seed = 0):
    """ Given data and label, split into test/train then give result """
    list_index = []
    X0 = np.arange(length)
    if k_split < 1:
        if label is None:
            label = np.ones(length)
        test_data, train_data, _, _ = train_test_split(pd.DataFrame(X0), label,
                                                       test_size    = k_split,
                                                       stratify     = label,
                                                       random_state = random_seed)
        split_values = (np.array(train_data.index), np.array(test_data.index))
        list_index.append(split_values)

    elif k_split == 1:
        split_values = (X0, X0)
        list_index.append(split_values)

    else:
        skf = StratifiedKFold(n_splits     = int(k_split),
                              random_state = random_seed,
                              shuffle      = True)
        for split_values in skf.split(X0, label):
            list_index.append(split_values)

    return list_index


def filter_gene(data, masks = None, index_gene = None, masking = False, scale_feature = False, use_3D = False):
    """ """
    if masks is None:
        if index_gene is None:
            raise Exception('Eitehr mask or index should be specified')
        else:
            masks = np.array([np.isin(np.arange(data.shape[-1]), index) for index in index_gene])

    if scale_feature is None:
        scale_feature = self.scale_feature

    if len(data.shape) == 3:

        assert masks.shape[0] == 1, "If data is 3D, then masks should only have one mask instead of {}".format(masks.shape[0])

        data = data.take(np.where(masks[0])[0], axis = -1)
        data = np.array([scale_float64(data_temp,0) for data_temp in data])

    else:
        print("mask shape: {}".format(masks.shape))
        if use_3D and not np.all(masks.sum(1) == masks.sum(1)[0]):
            raise Exception('For use_3D, all masks should contain same number of non-masked value')

        if masking:
            data = np.array(data)
        else:
            print(masks.sum())
            print(masks.shape)
            data = np.array([data[:,mask] for mask in masks])

        if scale_feature:
            data = np.array([scale_float64(data_masked,0) for data_masked in data])

        if not use_3D:
            assert len(data) == 1, "For use_3D is false, length of masks should be 1"
            data = data[0]

    return np.array(data)


def split_data(data_list, split_idx, axis_list = None):
    """ Split list of data by train/test index for specificed axis"""
    train_index, test_index = split_idx
    dataset = []
    #
    if axis_list is None:
        axis_list = [-2] * len(data_list)
    #
    for data,axis in zip(data_list, axis_list):
        dataset.append(data.take(train_index, axis))
        dataset.append(data.take(test_index, axis))
    #
    return dataset



#!/usr/bin/env python

from distutils.core import setup
from distutils.extension import Extension
from setuptools import find_packages

import numpy

## Snip from setup.py for tensorflow-forward-ad
try:
  from Cython.Distutils import build_ext
except ImportError:
  use_cython = False
else:
  use_cython = True

cmdclass = {}
ext_modules = []
include_dirs = [numpy.get_include()]

# if use_cython:
#   ext_modules += [
#       Extension(
#           "tensorflow_forward_ad.cbfs", ["tensorflow_forward_ad/cbfs.pyx"],
#           include_dirs=include_dirs),
#   ]
#   cmdclass.update({'build_ext': build_ext})
# else:
#   ext_modules += [
#       Extension(
#           "tensorflow_forward_ad.cbfs", ["tensorflow_forward_ad/cbfs.c"],
#           include_dirs=include_dirs),
#   ]


setup(name='siVAE',
      version='0.2',
      description='scalable and interpretable Variational Autoencoder',
      url='https://github.com/quon-titative-biology/siVAE',
      author=['Yongin Choi', 'Gerald Quon'],
      author_email='yonchoi@ucdavis.edu',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'pandas',
            'matplotlib',
            'scikit-learn',
            'seaborn',
            'tensorflow==1.15',
            "tensorflow-probability==0.8.0",
            'scipy',
            'scikit-image',
            'scanpy',
            'gseapy'],
      extras_requirements = [],
      cmdclass=cmdclass,
      ext_modules=ext_modules
      )



import os
import time
import logging

import tensorflow as tf

import numpy as np


def print_losses(losses, losses_name):
    """ Print losses in a specific format """
    zipped = zip(losses_name, losses)
    format_args = [item for pack in zipped for item in pack]
    str_report = "Train: {}={:.4}" + ", {}={:.4}" * (len(losses_name)-1)
    logging.info(str_report.format(*format_args))


def scope_name(scope):
    """ Returns scope name """

    scope_valid = False
    count = 0
    root = scope

    while not scope_valid:
        theta = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                                  scope = os.path.join(tf.get_variable_scope().name,scope))
        if len(theta) == 0:
            scope_valid = True
        else:
            count += 1
            scope = '{}_{}'.format(root, count)

    return scope


def reshape_sum(X, ImageDims):
    X_shape = X.shape[:-1]
    X_image = X.reshape(*X_shape,*ImageDims)
    X_image = X_image.sum(-1)
    X_image = X_image.reshape(*X_shape,*ImageDims[:-1],1)
    return X_image


class batch_handler():
    """ Handles mini-batches during model training """

    def __init__(self, total_size, mb_size):
        self.total_size = total_size
        self.idx_all = np.random.permutation(np.array(range(total_size)))
        self.i_epoch = 0
        self.mb_size = mb_size

    def next_batch(self):
        # input: idx_all, num, i_epoch
        # output: idx, ranom selection of the data
        # use in loop while updating idx_all and i_epoch
        start = self.i_epoch
        num = self.mb_size
        total_size = self.total_size
        idx_all = self.idx_all

        end = start + num
        new_i_epoch = end % total_size

        if end > total_size:
            idx_before_shuffle = idx_all[start:]
            idx_all = np.random.permutation(idx_all)
            end = new_i_epoch
            idx_after_shuffle = idx_all[:end]
            idx_out = np.concatenate((idx_before_shuffle, idx_after_shuffle))
        else:
            idx_out = idx_all[start:end]

        self.idx_all = idx_all
        self.i_epoch = new_i_epoch

        return idx_out



import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.ops import gen_nn_ops


# @ops.RegisterGradient("MaxPoolWithArgmax")
# def _MaxPoolGradWithArgmax(op, grad, unused_argmax_grad):
#   """The gradients for `MaxPoolWithArgmax`.
#   Args:
#     op: The `MaxPoolWithArgmax` `Operation` that we are differentiating, which
#       we can use to find the inputs and outputs of the original op.
#     grad: Gradient with respect to the output of the `MaxPoolWithArgmax` op.
#     op.inputs[0]: x
#     op.outputs[0]: y
#     op.outputs[1]: argmax_in_x
#   Returns:
#     Gradients with respect to the input of `MaxPoolWithArgmax`.
#   """
# 
#   return gen_nn_ops._max_pool_grad_with_argmax(
#       op.inputs[0],
#       grad,
#       op.outputs[1],
#       op.get_attr("ksize"),
#       op.get_attr("strides"),
#       padding=op.get_attr("padding"))


def _max_pool_grad_grad(dy, x, y, ksize, strides, padding, argmax=None):
  """Gradients of MaxPoolGrad."""
  if argmax is None:
    _, argmax = tf.nn.max_pool_with_argmax(x, ksize, strides, padding)
  grad = dy
  grad_flat = tf.reshape(grad, [-1])
  argmax_flat = tf.reshape(argmax, [-1])

  x_shape = tf.cast(tf.shape(x), argmax.dtype)
  batch_dim = tf.reshape(
      tf.range(
          x_shape[0], dtype=argmax.dtype), [-1, 1, 1, 1])
  nelem = tf.reduce_prod(x_shape[1:])
  batch_dim *= nelem

  y_zero = tf.zeros_like(y, dtype=argmax.dtype)
  batch_dim += y_zero
  batch_dim = tf.reshape(batch_dim, [-1])

  argmax_flat += batch_dim
  grad_input = tf.gather(grad_flat, argmax_flat)
  grad_input = tf.reshape(grad_input, tf.shape(y))
  return grad_input


@ops.RegisterGradient("MaxPoolGradWithArgmax")
def _MaxPoolGradWithArgmaxGrad(op, grad):
  """The gradients for `MaxPoolGradWithArgmax`.
  Args:
    op: The `MaxPoolGradWithArgmax` `Operation` that we are differentiating,
      which we can use to find the inputs and outputs of the original op.
    grad: Gradient with respect to the output of the `MaxPoolGradWithArgmax` op.
    op.inputs[0]: x
    op.inputs[1]: dl/dy
    op.inputs[2]: argmax_in_x
    op.outputs[0]: dl/dx
  Returns:
    Gradients with respect to the input of `MaxPoolGradWithArgmax`.
  """
  ksize = op.get_attr("ksize")
  strides = op.get_attr("strides")
  padding = op.get_attr("padding")
  return [
      None, _max_pool_grad_grad(
          grad,
          op.inputs[0],
          op.inputs[1],
          ksize,
          strides,
          padding,
          argmax=op.inputs[2]), None
  ]


# @ops.RegisterGradient("MaxPoolGrad")
# def _MaxPoolGradGrad(op, grad):
#   """The gradients for `MaxPoolGrad`.
#   Args:
#     op: The `MaxPoolGrad` `Operation` that we are differentiating, which we can use
#       to find the inputs and outputs of the original op.
#       op.inputs[0]: x
#       op.inputs[1]: y
#       op.inputs[2]: dl/dy
#       op.outputs[0]: dl/dx
#     grad: Gradient with respect to the output of the `MaxPoolGrad` op.
#   Returns:
#     Gradients with respect to the input of `MaxPoolGrad`.
#   """
#   ksize = op.get_attr("ksize")
#   strides = op.get_attr("strides")
#   padding = op.get_attr("padding")
#   return [
#       None, None, _max_pool_grad_grad(grad, op.inputs[0], op.inputs[1], ksize,
#                                       strides, padding)
#   ]



import os
import time
import logging

# Tensorflow
import tensorflow as tf
import tensorflow_probability as tfp

import numpy as np
import pandas as pd
import math

from .layer import layer
from .layer import KL_divergence_normal
from .layer import create_variance
from .layer import Distribution
from .util import batch_handler
from . import VAE
from .analysis import run_FI

from siVAE import FeatureImportance as FI
from siVAE import util

import scanpy as sc


def create_placeholder_W(VAE_sample):
    """
    Input: self.VAE_sample
    Return: placeholder (matrix of zeros) with
    """

    preW = VAE_sample.W
    preW_shape = [int(s) for s in preW.shape]
    preW_placeholder = np.zeros(preW.shape).transpose()

    return preW_placeholder


def setup_siVAE(datah_in, datah_in_transpose, graph_args1, graph_args2, graph_args3,
                graph = None, rs = 1):
    """ Set up siVAE """

    times = {}
    duration = {}
    times['start'] = time.time()

    if graph is None:
        graph = tf.get_default_graph()

    VAE_sample = VAE.AutoEncoder(data_handler = datah_in,
                                 random_seed  = rs,
                                 isVAE        = True,
                                 name         = 'VAE_sample',
                                 **graph_args1)
    VAE_sample.build_model(reset_graph = False, graph = graph)
    times['VAE_sample_build'] = time.time()
    ## Set datah_in_transpose for VAE_feature
    preW = VAE_sample.W
    preW_shape = [int(s) for s in preW.shape]
    preW_placeholder = np.zeros(preW.shape).transpose()

    datah_in_transpose.y = sc.AnnData(preW_placeholder)
    k_split = 1
    datah_in_transpose.create_split_index_list(k_split, random_seed = rs)
    ks = 0
    datah_in_transpose.create_dataset(kfold_idx = ks)

    ## Create VAE_feature
    VAE_feature = VAE.AutoEncoder(data_handler = datah_in_transpose,
                           random_seed = rs, isVAE = True,
                           name = 'VAE_feature', **graph_args2)

    VAE_feature.build_model(reset_graph = False, graph = graph)
    times['VAE_feature_build'] = time.time()

    VAE_dict = {'sample' : VAE_sample,
                'feature': VAE_feature}

    ## Create siVAE
    model = combinedVAE(random_seed = rs, name = 'combinedModel', **graph_args3)
    model.combine_models(VAE_dict)
    times['VAE_feature_build'] = time.time()

    return model, VAE_sample, VAE_feature


def run(datah_in, datah_in_transpose, graph_args1, graph_args2, graph_args3,
        rs = 1, do_pretrain = True,
        sample_input = None, do_FA = False, method_DE = [], kwargs_FI = {}):

    """ run """

    start_time = time.time()
    times = {}
    times['start'] = time.time()

    if datah_in.iterator is None:
        tf.reset_default_graph()
    graph = tf.get_default_graph()

    config = graph_args1['config']

    with tf.Session(config = config) as sess:

        with FI.DeepExplain(session=sess) as de:

            model, VAE_sample, VAE_feature = setup_siVAE(datah_in, datah_in_transpose,
                                            graph_args1, graph_args2, graph_args3,
                                            rs = rs, graph = graph)

            times['build_model'] = time.time()

            ## Train
            results1 = None
            results2 = None

            logdir_tf = graph_args1['logdir_tf']

            ## Initialize
            tf.set_random_seed(VAE_sample.random_seed)
            tf.logging.set_verbosity(tf.logging.FATAL)
            sess.run(tf.global_variables_initializer())

            var_init = sess.run([VAE_sample.X_dist.var, VAE_sample.z_dist.var],
                                feed_dict = {VAE_sample.X: VAE_sample.data_handler.X.X})

            ## Initialize tensorboard
            logdir_train = os.path.join(logdir_tf,'train')
            logdir_test  = os.path.join(logdir_tf,'test')
            train_writer = tf.summary.FileWriter(logdir_train, sess.graph)
            test_writer  = tf.summary.FileWriter(logdir_test, sess.graph)

            if do_pretrain:

                ## Train VAE_sample
                logging.info("Pre-train sample-wise encoder and decoder")
                results1 = VAE_sample.train(sess = sess,
                                            initialize = False,
                                            train_writer = train_writer,
                                            test_writer = test_writer,
                                            close_writer = False)

                times['train_VAE_sample'] = time.time()
                VAE_sample.sess  = sess

                ## Train VAE_feature
                logging.info("Pre-train feature-wise encoder and decoder")
                preW = results1['W'].transpose()
                VAE_feature.data_handler.y = sc.AnnData(preW)
                ks = 0
                VAE_feature.data_handler.create_dataset(kfold_idx = ks)

                results2 = VAE_feature.train(sess = sess,
                                             initialize = False,
                                             train_writer = train_writer,
                                             test_writer = test_writer,
                                             close_writer = False)
                times['train_VAE_feature'] = time.time()
                VAE_feature.sess = sess

            else:
                VAE_sample.sess  = sess
                VAE_feature.sess = sess
                times['train_VAE_sample']  = time.time()
                times['train_VAE_feature'] = time.time()

            ## Train model
            results = model.train(sess = sess, initialize = False, close_writer = True,
                                  train_writer = train_writer, test_writer = test_writer,
                                  pretrain = True)
            results = model.train(sess = sess, initialize = False, close_writer = True,
                                  train_writer = train_writer, test_writer = test_writer,
                                  pretrain = False)

            times['train_combined'] = time.time()

            results['duration'] = time.time() - start_time
            results['time']     = times

            ## Additional Analysis
            if sample_input is not None:

                sample_dict = {}

                ## Grid inputs -------------------------------------------------
                grids_list = []
                n_LE = int(model.VAE_sample.z_mu.shape[1])

                for ii in range(n_LE):
                    n_sample = 100
                    grids = np.zeros([n_sample,n_LE])
                    grids[:,ii] = np.arange(-1,1,0.02)
                    grids_list.append(grids)
                grids = np.concatenate(grids_list)
                X_dummy = np.zeros([len(grids),int(model.VAE_sample.X_target.shape[1])])

                feed_dict_full = {model.VAE_sample.X  : X_dummy,
                                  model.z_sample      : grids,
                                  model.X_target      : X_dummy,
                                  model.VAE_feature.X : model.VAE_dict['feature'].data_handler.X.X}

                sample_output = np.array(model.reconstruct_X(feed_dict_full))

                sample_dict['grid'] = sample_output

                ## Calculate components with sample input ----------------------
                feed_dict_full = {model.VAE_sample.X  : sample_input,
                                  model.X_target      : sample_input,
                                  model.VAE_feature.X : model.VAE_dict['feature'].data_handler.X.X}

                z_mu          = model.VAE_sample.calculate_z_mu(feed_dict_full)
                sample_output = np.array(model.reconstruct_X(feed_dict_full))

                sample_dict["input"]  = sample_input
                sample_dict["output"] = sample_output

                sampled_outputs = np.array([model.reconstruct_X(feed_dict_full)])
                sampled_outputs = np.swapaxes(sampled_outputs,0,1) # swap 0:Sample 1: Image


                ## Decoder layers ----------------------------------------------

                decoder_layers_list = []
                for ii in range(len(sampled_outputs)):

                    sampled_output = sampled_outputs[ii,]

                    feed_dict_full = {model.VAE_sample.X        : sampled_output,
                                      model.VAE_sample.X_target : sampled_output,
                                      model.VAE_feature.X       : model.VAE_dict['feature'].data_handler.X.X}

                    z_mu = model.VAE_sample.calculate_z_mu(feed_dict_full)
                    X_mu = np.array(model.reconstruct_X(feed_dict_full)) # Don't sample X
                    decoder_layers     = model.VAE_sample.calculate_decoder_layers(feed_dict_full)
                    decoder_layers     = [z_mu] + decoder_layers
                    decoder_layers[-1] = X_mu
                    decoder_layers_list.append(decoder_layers)

                sample_dict['decoder_layers'] = decoder_layers_list

                ## Feature Attribution -----------------------------------------

                if do_FA and len(method_DE) > 0:

                    logging.info('Perform feature attribution')

                    feed_dict_full = {model.VAE_sample.X        : sample_input,
                                      model.VAE_sample.X_target : sample_input,
                                      model.VAE_feature.X       : model.VAE_dict['feature'].data_handler.X.X}

                    sample_z = model.VAE_sample.calculate_z_mu(feed_dict_full)

                    feed_dict_base = {model.VAE_sample.X_target : sample_input,
                                      model.VAE_feature.X       : model.VAE_dict['feature'].data_handler.X.X}

                    attributions_dict_all = {}

                    ## Encoder
                    attributions_dict = run_FI(input     = model.VAE_sample.X,
                                               target    = model.VAE_sample.z_mu,
                                               sample_DE = sample_input,
                                               de        = de,
                                               sess      = sess,
                                               method_DE = method_DE,
                                               feed_dict = feed_dict_base,
                                               kwargs_FI = kwargs_FI,
                                               mode      = 'reverse')

                    attributions_dict_all['encoder'] = attributions_dict

                    ## Decoder
                    kwargs_FI2 = dict(kwargs_FI)
                    _ = kwargs_FI2.pop('baseline',None)
                    attributions_dict = run_FI(input     = model.VAE_sample.z_sample,
                                               target    = model.X_dist.mu,
                                               sample_DE = sample_z,
                                               de        = de,
                                               sess      = sess,
                                               method_DE = method_DE,
                                               feed_dict = feed_dict_base,
                                               kwargs_FI = kwargs_FI2,
                                               mode      = 'forward')

                    attributions_dict_all['decoder'] = attributions_dict

                    sample_dict['attributions_samples'] = attributions_dict_all

                results['sample_dict'] = sample_dict

    return results, results1, results2, model


class combinedVAE(object):

    def __init__(self, logdir_tf = ".", data_handler = None, iter = 5000,
                 h_dims = None, LE_dim = None, architecture = None,
                 mb_size = 1000, learning_rate = 1e-3, l2_scale = 0.0, l1_scale = 0.0,
                 early_stopping = 0, tolerance = 0, min_early_stopping = 0,
                 activation_fun = tf.nn.relu, random_seed = 0,
                 log_frequency = 100, batch_norm = False, keep_prob = None, masking = False,
                 dataAPI = False, tensorboard = False, metadata = False, permute_axis = -2,
                 custom = False, config = None, validation_split = 0,
                 decay_rate=1, decay_steps = 1000, save_recon = True, save_LE = True,
                 decoder_var='scalar', save_W=True, set_y_logvar=False, decoder_activation = None,
                 name="", var_dependency=True, optimizer_type=tf.compat.v1.train.AdamOptimizer,
                 kernel_initializer=tf.contrib.layers.xavier_initializer(uniform = False), X_mu_use_bias=True,
                 zv_recon_scale=None, hl_recon_scale=None):

        self.decoder_var = decoder_var
        self.custom = custom
        self.var_dependency = var_dependency
        self.X_mu_use_bias = X_mu_use_bias
        self.decoder_activation = decoder_activation

        ## Recalculate h_dims if h_dims = 0
        self.save_LE = save_LE
        self.save_W = save_W
        self.set_y_logvar = set_y_logvar
        self.save_recon = save_recon

        ## Variable scope name
        self.name = name

        ## Set activation function
        self.activation_fun = activation_fun

        ## Training Parameters
        self.mb_size = mb_size
        self.iter = iter
        self.keep_prob = keep_prob

        ## Hyperparameters
        self.random_seed = random_seed

        # learning rate
        self.learning_rate = learning_rate
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate

        # Regularization
        self.l1_scale = l1_scale
        self.l2_scale = l2_scale
        self.batch_norm = batch_norm
        self.placeholder_reg = {}

        ## Early stopping
        self.validation_split = validation_split
        self.tolerance = tolerance
        self.early_stopping = early_stopping
        self.min_early_stopping = min_early_stopping
        self.iter_min_early_stopping = self.iter * self.min_early_stopping

        ## Tensorboard
        self.tensorboard = tensorboard
        self.logdir_tf = logdir_tf
        self.metadata = metadata
        self.summary_list = []

        self.feed_dict = {}

        if self.metadata and not self.tensorboard:
            logging.info("Tensorboard has been switched to true as metadata requires tensorboard")
            self.tensorboard = True

        self.log_frequency = log_frequency
        self.permute_axis = permute_axis
        self.dataAPI = dataAPI
        self.config = config


    def combine_models(self, VAE_dict):
        """

        """

        self.VAE_dict = VAE_dict

        assert all(VAE_type in list(VAE_dict.keys())  for VAE_type in ['sample', 'feature']), 'VAE_dict must contain both cell and gene type'

        self.VAE_sample  = self.VAE_dict['sample']
        self.VAE_feature = self.VAE_dict['feature']
        self.y_var_type  = self.VAE_sample.y_var_type
        self.do_sample   = self.VAE_sample.do_sample
        self.X_target    = self.VAE_sample.X_target
        self.decoder_var = self.VAE_sample.decoder_var
        self.isVAE       = self.VAE_sample.isVAE

        with tf.variable_scope(self.name, reuse = tf.AUTO_REUSE):

            self.v_sample = self.VAE_feature.z_sample
            self.v_mu     = self.VAE_feature.z_mu
            self.v_sample = self.v_mu

            self.z_sample_pre = self.VAE_sample.z_sample
            self.z_mu_pre     = self.VAE_sample.z_mu


            ## Combine with feature embeddings, v_sample
            self.feature_weights = tf.get_variable(name = 'feature_weights',
                                                   shape = [int(self.VAE_sample.X.shape[-1]),int(self.v_sample.shape[-1])],
                                                   dtype = tf.float32,
                                                   initializer = tf.keras.initializers.RandomNormal(),
                                                   trainable = True)

            self.summary_list.append(tf.summary.histogram('Feature Weights',self.feature_weights))
            self.summary_list.append(tf.summary.histogram('Feature Weights',self.feature_weights))
            self.summary_list.append(tf.summary.histogram('Feature Weights',self.feature_weights))

            self.weighted_v = self.feature_weights * self.v_sample
            self.weighted_v_sum = tf.math.reduce_mean(self.weighted_v,axis=0) + 1

            self.z_sample = self.z_sample_pre * self.weighted_v_sum
        print('VAE_sample decoder variable scope')
        print(self.VAE_sample.decoder_variable_scope)
        with tf.variable_scope(self.VAE_sample.decoder_variable_scope, reuse = True):

            self.build_decoder(h_temp   = self.z_sample,
                               h_dims   = self.VAE_sample.h_dims_decoder,
                               l1_scale = self.VAE_sample.l1_scale,
                               l2_scale = self.VAE_sample.l2_scale,
                               variable_scope = 'Decoder',
                               kernel_initializer = self.VAE_sample.kernel_initializer)

        self.create_decoder_loss()
        self.create_loss()

        self.summary_list += self.VAE_sample.summary_list


    def build_decoder(self, h_dims, h_temp, l1_scale, l2_scale, kernel_initializer, variable_scope = 'decoder'):
        """
        Build decoder
        """
        logging.info("Building " + variable_scope + "with dims {}".format(h_dims))

        self.decoder_layers = []
        self.hidden_layers  = []

        with tf.variable_scope(variable_scope, reuse = tf.AUTO_REUSE):

            self.decoder_variable_scope = tf.get_variable_scope().name

            for ii, h_dim in enumerate(h_dims):

                with tf.variable_scope("hidden_layer_{}".format(ii)):

                    if ii < len(h_dims) - 2:

                        ## Hidden layers
                        fun = self.activation_fun

                        h_temp,_,_ = layer(h = h_temp, dim = h_dim, fun = fun,
                                           l2_scale = l2_scale, l1_scale = l1_scale,
                                           kernel_initializer = kernel_initializer)

                    elif ii == len(h_dims) - 2:
                        ## Last layer before final layer

                        prob_layer  = Distribution(h   = h_temp,
                                                   dim = h_dim,
                                                   fun = None,
                                                   var_type = self.y_var_type,
                                                   l2_scale = l2_scale,
                                                   l1_scale = l1_scale)

                        self.y_dist = prob_layer
                        self.y_sample, self.y_mu, self.y_var, self.y_logvar, self.y_eps = prob_layer.get_tensors()
                        h_temp = self.y_sample

                    elif ii == len(h_dims) - 1:

                        ## If there is no y layer, set y to z
                        if len(h_dims) == 1:
                            self.y_dist = self.z_dist
                            self.y_sample = self.z_sample
                            self.y_mu = self.z_mu
                            self.y_var = self.z_var
                            self.y_logvar = self.z_logvar
                            self.y_eps    = self.z_eps

                        with tf.variable_scope('output', reuse = tf.AUTO_REUSE):

                            self.output_variable_scope = tf.get_variable_scope().name

                            prob_layer,W,b = self.build_output_layer(h_temp, h_dim = h_dim,
                                                                     fun = self.activation_fun,
                                                                     var_type = self.decoder_var,
                                                                     var_dependency = self.var_dependency,
                                                                     l2_scale = l2_scale,
                                                                     l1_scale = l1_scale,
                                                                     use_bias = self.X_mu_use_bias)

                        self.X_dist = prob_layer
                        self.X_sample, self.X_mu, self.X_var, self.X_logvar, self.X_eps = prob_layer.get_tensors()
                        self.X_sample = tf.cond(self.do_sample, lambda: self.X_sample, lambda: self.X_mu, 'X_sample_cond')
                        self.W = W
                        self.b = b
                        h_temp = self.X_mu

                self.decoder_layers.append(h_temp)
                self.hidden_layers.append(h_temp)
                self.summary_list.append(tf.summary.histogram("hidden_layer_{}", h_temp))


        return prob_layer


    def build_output_layer(self, h_temp, fun, var_type, var_dependency, l2_scale, l1_scale, use_bias,
                           h_dim = None, W = None, b = None, h_mu = None, h_std = None, custom = False, **kwargs):
        """"""

        h_mu, W, b = layer(h_temp, h_dim, fun = None,
                           name = "{}_{}".format('X','mu'),
                           l1_scale = l1_scale, l2_scale = l2_scale,
                           custom = custom, use_bias = use_bias,
                           W = W, b = b, **kwargs)

        with tf.variable_scope('Distribution'):
            prob_layer  = Distribution(h = h_temp,
                                       h_mu = h_mu,
                                       h_std = h_std,
                                       dim = h_dim,
                                       fun = None,
                                       var_type = var_type,
                                       var_dependency = var_dependency,
                                       l2_scale = l2_scale,
                                       l1_scale = l1_scale,
                                       **kwargs)

        return prob_layer, W, b


    def mean_squared_error(self, target, reconstruction):
        """ calculate mean squared error between target and reconstruction """
        return tf.square(target - reconstruction)


    def create_decoder_loss(self):
        """ Create decoder loss """

        with tf.name_scope('mean_squared_error'):
            self.recon_loss_ae = self.mean_squared_error(self.X_mu, self.X_target)

        with tf.name_scope('log_likelihood'):
            if self.isVAE and self.decoder_var != 'deterministic':
                self.decoder_loss_old        = self.X_logvar / 2 + self.recon_loss_ae / self.X_var / 2 + tf.constant(0.5 * np.log(2. * np.pi), dtype = self.X_var.dtype)
                self.decoder_loss_mse        = tf.square(self.X_mu - self.X_target)
                self.decoder_loss_mse_scaled = self.recon_loss_ae / self.X_var / 2
                self.decoder_loss_logvar     = self.decoder_loss_mse * 0 + self.X_logvar / 2
                self.decoder_loss            = -self.X_dist.dist.log_prob(self.X_target)
            else:
                self.decoder_loss = tf.reduce_sum(self.recon_loss_ae,-1)


    def create_loss(self):
        """
        Crate losses
        """

        ## l2_loss
        self.reg_loss = tf.reduce_sum(tf.losses.get_regularization_losses(scope = self.name))
        logging.info(" ")
        logging.info("reg_loss: {}".format(self.reg_loss.shape))
        logging.info(self.reg_loss)
        logging.info(tf.losses.get_regularization_losses(scope = self.name))

        ## recon_loss
        self.recon_loss_scalar = tf.reduce_mean(tf.reduce_sum(self.recon_loss_ae,-1))
        logging.info("recon_loss_scalar: {}".format(self.recon_loss_scalar.shape))
        logging.info(self.recon_loss_scalar)
        logging.info(self.recon_loss_ae)

        ## decoder_loss
        self.decoder_loss_scalar = tf.reduce_mean(self.decoder_loss)
        tf.losses.add_loss(self.decoder_loss_scalar)
        logging.info("decoder_loss_scalar: {}".format(self.decoder_loss_scalar.shape))
        logging.info(self.decoder_loss_scalar)
        logging.info(self.decoder_loss)

        ## Total loss consists of KL
        self.total_loss = self.VAE_sample.KL_loss_scalar + \
                          self.VAE_feature.KL_loss_scalar + \
                          self.decoder_loss_scalar + \
                          self.VAE_sample.reg_loss + \
                          self.VAE_feature.reg_loss + \
                          self.reg_loss

        self.losses = [self.total_loss,
                       self.decoder_loss_scalar,
                       self.recon_loss_scalar,
                       self.VAE_sample.reg_loss,
                       self.VAE_feature.reg_loss,
                       self.VAE_sample.KL_loss_scalar,
                       self.VAE_feature.KL_loss_scalar,
                       self.reg_loss]

        self.losses_name = ['total loss',
                            'decoder_loss',
                            'recon_loss',
                            'reg_loss sample',
                            'reg_loss feature',
                            'latent loss sample',
                            'latent loss feature',
                            'reg_loss']

        for loss_name, loss in zip(self.losses_name, self.losses):
            self.summary_list.append(tf.summary.scalar(loss_name, loss))

        ## Optimizer
        # Variables to Train
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS, scope=self.name)
        self.theta = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope = self.name)
        # self.theta += self.VAE_sample.theta + self.VAE_feature.theta
        # self.theta += self.VAE_sample.theta

        self.theta2 = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope = self.name)
        self.theta += self.VAE_sample.theta + self.VAE_feature.theta
        # self.theta2 += self.VAE_sample.theta

        ## Set up optimizer
        with tf.name_scope("global_steps"):

            self.global_step = tf.Variable(0, name='global_step', trainable=False)

            self.decayed_learning_rate = tf.train.exponential_decay(learning_rate = self.learning_rate,
                                                                    global_step   = self.global_step,
                                                                    decay_steps   = self.decay_steps,
                                                                    decay_rate    = self.decay_rate,
                                                                    staircase     = False)

        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        var_list = self.theta

        logging.info("Build train_op optimizing:")
        logging.info(var_list)

        with tf.name_scope("train_op"):
            self.optimizer = tf.train.AdamOptimizer(self.decayed_learning_rate)

            with tf.control_dependencies(update_ops):
                optimizer = self.optimizer.minimize(loss = self.total_loss,
                                                    global_step = self.global_step,
                                                    var_list = var_list)
                self.train_op = optimizer

        var_list = self.theta2
        with tf.name_scope("train_op2"):
            self.optimizer2 = tf.train.AdamOptimizer(self.decayed_learning_rate)

            with tf.control_dependencies(update_ops):
                optimizer = self.optimizer2.minimize(loss = self.total_loss,
                                                    global_step = self.global_step,
                                                    var_list = var_list)
                self.train_op2 = optimizer


        logging.info("train_op:")
        logging.info(self.train_op)

        if self.tensorboard:
            self.merge = tf.summary.merge(self.summary_list)
            self.losses.insert(0,self.merge)


    def start_tensorboard(self, train_writer = None, test_writer = None):
        """ Start tensorboard """

        if train_writer is not None and test_writer is not None:
            self.train_writer = train_writer
            self.test_writer  = test_writer
        else:
            logdir_train = os.path.join(self.logdir_tf,self.name,'train')
            logdir_test  = os.path.join(self.logdir_tf,self.name,'test')
            self.train_writer = tf.summary.FileWriter(logdir_train, self.sess.graph)
            self.test_writer = tf.summary.FileWriter(logdir_test, self.sess.graph)
            logging.info("Starting Tensorboard: {}".format(logdir_train))


    def train(self, sess, dataset = None, initialize = True, train_writer = None, test_writer = None, close_writer = True, pretrain = False):
        """ Train the model """

        start_time = time.time()

        self.sess = sess

        # Set random seed
        if initialize:
            tf.set_random_seed(self.random_seed)
            tf.logging.set_verbosity(tf.logging.FATAL)
            sess.run(tf.global_variables_initializer())

        dataset_sample = self.VAE_dict['sample'].data_handler.dataset
        train_data_full, test_data, y_train_full, y_test = dataset_sample
        dataset_feature = self.VAE_dict['feature'].data_handler.dataset

        # Split into validation
        if self.validation_split == 0:
            train_data = train_data_full
            y_train = y_train_full
            validation_data = test_data
            y_validation = y_test

        else:
            len_train = int(train_data.shape[self.permute_axis])
            len_validate = int(len_train * self.validation_split)
            validation_data = train_data_full.swapaxes(0,self.permute_axis)[:len_validate].swapaxes(0,self.permute_axis)
            train_data      = train_data_full.swapaxes(0,self.permute_axis)[len_validate:].swapaxes(0,self.permute_axis)
            y_validation    = y_train.swapaxes(0,self.permute_axis)[:len_validate].swapaxes(0,self.permute_axis)
            y_train         = y_train.swapaxes(0,self.permute_axis)[len_validate:].swapaxes(0,self.permute_axis)

        logging.info("======================== Data Split =============================")
        logging.info("validation_data: " + str(validation_data.shape))
        logging.info("y_validation: " + str(y_validation.shape))
        logging.info("train_data: " + str(train_data.shape))
        logging.info("y_train: " + str(y_train.shape))
        logging.info("test_data: " + str(test_data.shape))
        logging.info("y_test: " + str(y_test.shape))
        logging.info('')

        logging.info("======================== losses =============================")
        logging.info(self.losses)
        logging.info(self.losses_name)

        if self.tensorboard:
            "Starting Tensorboard "
            self.start_tensorboard(train_writer = train_writer, test_writer=  test_writer)

        feed_dict = {}
        feed_dict[self.VAE_feature.X] = dataset_feature[0]

        # Set up for early stopping
        it_tb = -1
        it_test = 0
        it_min = 0
        min_loss = float("inf")
        result = None

        feed_dict_train = {self.VAE_sample.X: train_data,
                           self.VAE_sample.X_target: y_train,
                           self.VAE_feature.X: dataset_feature[0]}

        feed_dict_test = {self.VAE_sample.X: validation_data,
                          self.VAE_sample.X_target: y_validation,
                          self.VAE_feature.X: dataset_feature[0]}

        iter = self.iter

        for VAE_in in self.VAE_dict.values():
            if VAE_in.keep_prob is not None:
                feed_dict[VAE_in.prob] = VAE_in.keep_prob

        ## Set minibatch size
        if self.mb_size <= 1:
            mb_size = int(int(train_data.shape[self.permute_axis]) * self.mb_size)
        else:
            mb_size = int(self.mb_size)

        train_data_size = int(train_data.shape[self.permute_axis])

        # if self.VAE_sample.iterator is None:
        batch = batch_handler(train_data_size, mb_size)

        # Set up batch
        if pretrain:
            self.solvers = [self.train_op2]
        else:
            self.solvers = [self.train_op]

        for it in range(iter):
            ## Separate counters for tensorboard and early stopping
            it_tb += 1
            it_test += 1

            idx_out = batch.next_batch()

            if self.VAE_sample.iterator is None:
                feed_dict[self.VAE_sample.X]        = train_data.take(idx_out, axis = -2)
                feed_dict[self.VAE_sample.X_target] = y_train.take(idx_out, axis = -2)

            ## Train a batch
            results = self.sess.run(self.solvers + self.losses, feed_dict = feed_dict)

            # Log train/test results
            if (it_tb % self.log_frequency == 0) or (it_tb == iter - 1):
                logging.info('')
                logging.info('Iter: {}, time: {:.4}s'.format(it, time.time()- start_time))

                ## Batch Loss
                zipped = zip(self.losses_name, results[-len(self.losses_name):])
                format_args = [item for pack in zipped for item in pack]
                str_report = "Batch: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                logging.info(str_report.format(*format_args))

                ## Train Loss
                results = sess.run(self.losses, feed_dict = feed_dict_train)
                results_train = results

                # Write to Summary (writing train data)
                if self.tensorboard:
                    summary = results.pop(0)
                    self.train_writer.add_summary(summary, str(it_tb))

                zipped = zip(self.losses_name,results[-len(self.losses_name):])
                format_args = [item for pack in zipped for item in pack]
                str_report = "Train: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                logging.info(str_report.format(*format_args))

                ## Validation loss

                if self.tensorboard:
                    run_args = {}

                    if self.metadata:
                        run_args["options"] = tf.RunOptions(trace_level=tf.RunOptions.FULL_TRACE)
                        run_args['run_metadata'] = tf.RunMetadata()

                    results = self.sess.run(self.losses,
                                       feed_dict = feed_dict_test, **run_args)

                    summary = results.pop(0)

                    if self.metadata:
                        self.test_writer.add_run_metadata(run_args["run_metadata"], str(it_tb))

                    self.test_writer.add_summary(summary, str(it_tb))

                    zipped = zip(self.losses_name,results[-len(self.losses_name):])
                    format_args = [item for pack in zipped for item in pack]
                    str_report = "Test: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                    logging.info(str_report.format(*format_args))

                else:
                    results = sess.run(self.losses, feed_dict = feed_dict_test)
                    zipped = zip(self.losses_name,results[-len(self.losses_name):])
                    format_args = [item for pack in zipped for item in pack]
                    str_report = "Test: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                    logging.info(str_report.format(*format_args))

                results_test = results

                if self.early_stopping > 0:
                    logging.info("new min loss: {} vs old {}".format(results[-3], min_loss))
                    delta = min_loss - results[-3]
                    if delta > self.tolerance:
                        it_min = it
                        it_test = 0
                        min_loss =  results[-3]

                    elif it_test > self.early_stopping:
                        break

        if self.tensorboard and close_writer:
            self.train_writer.close()
            self.test_writer.close()

        result = {}

        output = {'model_evaluation': result}
        output['iter'] = it
        output['losses'] = {'train': results_train,
                            'test' : results_test,
                            'name' : self.losses_name}

        #### Calculate layers for full data and save results
        output['split_index'] = self.VAE_dict['sample'].data_handler.split_idx
        output['labels']      = np.array(self.VAE_dict['sample'].data_handler.X.obs['Labels'])

        feed_dict_full = {self.VAE_sample.X        : self.VAE_dict['sample'].data_handler.X.X,
                          self.VAE_sample.X_target : self.VAE_dict['sample'].data_handler.y.X,
                          self.VAE_feature.X       : self.VAE_dict['feature'].data_handler.X.X}

        y_full    = self.VAE_dict['sample'].data_handler.y.X
        full_data = self.VAE_dict['sample'].data_handler.X.X

        if self.save_recon:

            logging.info("y_full: {}".format(y_full.shape))

            y_reconstruct = np.array(self.reconstruct_X(feed_dict_full))

            y_residual = y_full - y_reconstruct

            y_result = np.array([y_full, y_reconstruct])

            output['reconstruction'] = y_result

        if self.save_LE:

            bottleneck = {}
            for VAE_type, VAE_in in self.VAE_dict.items():
                bottleneck[VAE_type] = np.array(self.calculate_latent_embedding(VAE_in, feed_dict_full))

            output['latent_embedding'] = bottleneck

            bottleneck_var = {}
            for VAE_type, VAE_in in self.VAE_dict.items():
                bottleneck_var[VAE_type] = np.array(self.calculate_latent_variance(VAE_in, feed_dict_full))

            output['latent_embedding_var'] = bottleneck_var

        if self.save_W:
            W = self.get_W(feed_dict_full)
            output['W_mu'] = W
            y_mu = self.VAE_sample.calculate_y_mu(feed_dict_full)
            output['y_mu'] = y_mu

            weighted_v = self.get_feature_weights(feed_dict_full)
            output['feature_weights'] = weighted_v

            ## Decoder layers
            decoder_layers_dict = {'sample' : self.VAE_sample.calculate_decoder_layers(feed_dict_full),
                                   'feature': self.VAE_feature.calculate_decoder_layers(feed_dict_full)}
            output['decoder_layers'] = decoder_layers_dict

        return output


    def calculate_X_reconstruct(self, feed_dict_in):
        """ Reconstruct X from input data """

        y_reconstruct = np.array(self.sess.run(self.X_dist.mu, feed_dict = feed_dict_in))

        return y_reconstruct


    def calculate_latent_embedding(self, VAE_in, feed_dict_in):
        """ Calculate bottleneck layer from input data"""

        bottleneck = np.array(self.sess.run(VAE_in.latent_embedding, feed_dict = feed_dict_in))

        return bottleneck


    def reconstruct_X(self, feed_dict_in):
        """
        Calculate the X_reconstruct from the mean of the latent variables rather than sampled
        y_reconstruct = [cell, gene]
        """

        z_sample = self.calculate_latent_embedding(self.VAE_sample,feed_dict_in)
        v_sample = self.calculate_latent_embedding(self.VAE_feature,feed_dict_in)

        feed_dict_Wy = {self.v_sample: v_sample, self.z_sample_pre: z_sample}

        X_reconstruct = np.array(self.sess.run(self.X_dist.mu, feed_dict = feed_dict_Wy))

        return X_reconstruct


    def calculate_latent_variance(self, VAE_in, feed_dict_in):
        """ Calculate latent variance of VAE_in """
        bottleneck = np.array(self.sess.run(VAE_in.z_var, feed_dict = feed_dict_in))

        return bottleneck


    def calculate_decoder_variance(self, test_data):
        """ Calculate decoder variance """
        var = np.array(self.sess.run(self.X_var, feed_dict = {self.X: test_data}))

        return var


    def get_W(self, feed_dict = {}):
        """ Get W"""
        W = np.array(self.sess.run(self.W, feed_dict=feed_dict))

        return W


    def get_feature_weights(self, feed_dict = {}):
        """ Get W"""
        W = np.array(self.sess.run(self.feature_weights, feed_dict=feed_dict))

        return W



import os
import time
import logging

from deepexplain.tensorflow import DeepExplain
from tensorflow_forward_ad import forward_gradients

import numpy as np

import tensorflow as tf

def DE(input, target, sample, method, de, mask = None,  feed_dict = {}, **kwargs):
    """

    Parameters
    ----------
    input
        Input tensor
    target
        Target tensor
    sample
        Sample numpy array
    de
        deepexplain object
    mask
        mask the length of target tensor
    method_DE
        List of methods for feature attribution
        ['Saliency Maps', 'Gradient * Input', 'Integrated Gradients', 'DeepLIFT', 'Epsilon-LRP', 'Occlusion']
    feed_dict
        Additional feed_dict arguments

    Return
    ------
    Dictionary that maps method to feature attribution that is [sample x input]
    """

    dict_DE = {'Saliency Maps'       : 'saliency',
               'Gradient * Input'    : 'grad*input',
               'Integrated Gradients': 'intgrad',
               'Epsilon-LRP'         : 'elrp',
               'DeepLIFT'            : 'deeplift',
               'Occlusion'           : 'occlusion',
               'Shapley Value'       : 'shapley_sampling'}

    if len(target.shape) == 0:
        mask = 1
    else:
        if mask is None:
            mask = np.ones(target.shape[1])

    attributions = {}

    kwargs_in = {}
    if method in ['DeepLIFT','Integrated Gradients']:
        if 'baseline' in kwargs.keys():
            kwargs_in['baseline'] = kwargs['baseline']
    elif method in ['Integrated Gradients']:
        if 'steps' in kwargs.keys():
            kwargs_in['steps'] = kwargs['steps']
    elif method in ['DeepLIFT']:
        if 'init_ref' in kwargs.keys():
            kwargs_in['init_ref'] = kwargs['init_ref']

    attributions = de.explain(method = dict_DE[method],
                              T = target * mask,
                              X = input,
                              xs = sample,
                              feed_dict = feed_dict,
                              **kwargs_in)
    return attributions


def DE_marginal(input, target, sample, method, de, masks = None, feed_dict = {}, **kwargs):
    """
    Parameters

    ----------
    input
        Input tensor
    target
        Target tensor
    sample
        Sample numpy array
    method
        method for feature attribution
        ['Saliency Maps', 'Gradient * Input', 'Integrated Gradients', 'DeepLIFT', 'Epsilon-LRP', 'Occlusion']
    de
        deepexplain object
    masks
        List of mask the length of target tensor
    feed_dict
        Additional feed_dict arguments
    """

    dict_DE = {'Saliency Maps'       : 'saliency',
               'Gradient * Input'    : 'grad*input',
               'Integrated Gradients': 'intgrad',
               'Epsilon-LRP'         : 'elrp',
               'DeepLIFT'            : 'deeplift',
               'Occlusion'           : 'occlusion',
               'Shapley Value'       : 'shapley_sampling'}

    marginal = []

    if len(target.shape) in [0,1]:
        raise Exception('Shape of target must be [num_sample,num_features]')
    else:
        n_node = int(target.shape[1])

    if masks is None:
        masks = [np.arange(n_node) == ii for ii in range(n_node)]
    elif masks == 'sum':
        masks = [np.ones(n_node)]

    kwargs_in = {}
    if method in ['DeepLIFT','Integrated Gradients']:
        if 'baseline' in kwargs.keys():
            kwargs_in['baseline'] = kwargs['baseline']
    elif method in ['Integrated Gradients']:
        if 'steps' in kwargs.keys():
            kwargs_in['steps'] = kwargs['steps']
    elif method in ['DeepLIFT']:
        if 'init_ref' in kwargs.keys():
            kwargs_in['init_ref'] = kwargs['init_ref']

    explainer = de.get_explainer(method = dict_DE[method],
                                 T = target,
                                 X = input,
                                 feed_dict = feed_dict,
                                 **kwargs_in)

    for mask in masks:
        attributions = explainer.run(xs = sample,
                                     ys = [mask] * len(sample))
        marginal.append(attributions)
        # marginal.append(DE(input     = input,
        #                    target    = target,
        #                    sample    = sample,
        #                    de        = de,
        #                    mask      = mask,
        #                    method    = method,
        #                    feed_dict = feed_dict,
        #                    **kwargs))

    marginal = np.array(marginal)        # [n_target, n_sample, n_input]
    marginal = np.moveaxis(marginal,0,2) # [n_sample, n_input, n_target]

    return(marginal)


def FG(input, target, sample, method, sess, masks = None, feed_dict = {}, **kwargs):
    """
    Parameters
    ----------
    input
        Input tensor
    target
        Target tensor
    sample
        Sample numpy array
    de
        deepexplain object
    masks
        List of mask the length of target tensor
    method_DE
        List of methods for feature attribution
        ['Saliency Maps', 'Gradient * Input']
    feed_dict
        Additional feed_dict arguments
    """

    if len(input.shape) in [0,1]:
        raise Exception('Shape of target must be [num_sample,num_features]')
    else:
        n_node = int(input.shape[1])

    if masks is None:
        masks = [np.arange(n_node) == ii for ii in range(n_node)]
    elif masks == 'sum':
        masks = [np.ones(n_node)]

    v = tf.placeholder(tf.float32,[1,int(input.shape[1])])

    dydx = forward_gradients(target,input,v)[0]
    if method == 'Saliency Maps':
        FA = dydx
    elif method == 'Gradient * Input':
        FA = dydx * target
    else:
        raise Exception('Invalid feature attribution method')

    marginal = []

    for mask in masks:
        feed_dict_in = feed_dict.copy()
        feed_dict_in[input] = sample
        feed_dict_in[v]     = mask.reshape(1,-1)
        attributions = sess.run(FA, feed_dict = feed_dict_in)

        marginal.append(attributions)

    marginal = np.array(marginal)        # [n_input, n_sample, n_target]
    marginal = np.moveaxis(marginal,0,1) # [n_sample, n_input, n_target]

    return(marginal)


def feature_attribution(input, target, sample, sess = None, de = None, masks = None, method_DE = ['Saliency Maps'], feed_dict = {}, mode = 'forward', **kwargs):
    """

    Parameters
    ----------
    input
        Input tensor
    target
        Target tensor
    sample
        Sample numpy array
    de
        deepexplain object
    masks
        List of mask the length of target tensor
    method_DE
        List of methods for feature attribution
        ['Saliency Maps', 'Gradient * Input', 'Integrated Gradients', 'DeepLIFT', 'Epsilon-LRP', 'Occlusion']
    feed_dict
        Additional feed_dict arguments

    Return
    ------
    attributions
        [ n_method x n_sample x n_input x n_target ]
    """

    forward_mode_methods = ['Saliency Maps', 'Gradient * Input']

    attributions = []

    for method in method_DE:

        # if mode == 'forward' and method in forward_mode_methods:
        #     attribution = FG(input, target, sample, method, sess, masks = None, feed_dict = feed_dict, **kwargs)
        # else:
        attribution = DE_marginal(input, target, sample, method, de, masks = None, feed_dict = feed_dict, **kwargs)

        attributions.append(attribution)

    attributions = np.array(attributions)

    return(attributions)
